import 'dart:convert';
import 'dart:developer';
import 'package:flutter/cupertino.dart';
import 'package:nui_core/code/nui_code_util.dart';
import 'package:nui_core/datetime/nui_dt_util.dart';
import 'package:nui_core/string/nui_string_util.dart';

import 'nui_data_util.dart';

typedef EntFieldSetter<T> = Object Function(T data, dynamic value);
typedef EntFieldGetter<T> = dynamic Function(T data);
typedef EntityGetter<T> = NUIEntMapper<T> Function();
typedef EntCustomerMapper = dynamic Function(dynamic value);

enum NUIEntDisplayType{
  LABEL,
  NAME,
  DATE,
  TIME,
  CURRENCY,
  DISTANCE,
  ADDRESS,
  IMAGE_URL,
  IMAGE_FILE,
  COLOR
}

enum NUIEntType{
  STRING,
  INTEGER,
  DOUBLE,
  DATE,
  BOOLEAN,
  OBJECT,
  MAP,
  LIST_INT,
  LIST_STRING,
  LIST_DOUBLE,
  LIST_OBJECT,
  LIST_MAP,
}

NUIEntType valueOfEntType(String value){
  switch(value){
    case "STRING" : return NUIEntType.STRING;
    case "INTEGER" : return NUIEntType.INTEGER;
    case "DOUBLE" : return NUIEntType.DOUBLE;
    case "DATE" : return NUIEntType.DATE;
    case "BOOLEAN" : return NUIEntType.BOOLEAN;
    case "OBJECT" : return NUIEntType.OBJECT;
    case "MAP" : return NUIEntType.MAP;
    case "LIST INT" : return NUIEntType.LIST_INT;
    case "LIST STRING" : return NUIEntType.LIST_STRING;
    case "LIST DOUBLE" : return NUIEntType.LIST_DOUBLE;
    case "LIST OBJECT" : return NUIEntType.LIST_OBJECT;
    case "LIST MAP" : return NUIEntType.LIST_MAP;
    default: return null;
  }
}

extension NUIEntTypeExt on NUIEntType{
  String value(){
    switch(this){
      case NUIEntType.STRING : return "STRING";
      case NUIEntType.INTEGER : return "INTEGER";
      case NUIEntType.DOUBLE : return "DOUBLE";
      case NUIEntType.DATE : return "DATE";
      case NUIEntType.BOOLEAN : return "BOOLEAN";
      case NUIEntType.OBJECT : return "OBJECT";
      case NUIEntType.MAP : return "MAP";
      case NUIEntType.LIST_INT : return "LIST INT";
      case NUIEntType.LIST_STRING : return "LIST STRING";
      case NUIEntType.LIST_DOUBLE : return "LIST DOUBLE";
      case NUIEntType.LIST_OBJECT : return "LIST OBJECT";
      case NUIEntType.LIST_MAP : return "LIST MAP";
      default: return null;
    }
  }
}

class NUIFormatting{
  NUIFormatting._builder();
  static const String TWO_DECIMAL = "TWO DECIMAL";
  static const String ONE_DECIMAL = "ONE DECIMAL";
  static const String ROUNDED = "ROUNDED";
}

D asObject<D>(NUIEntMapper<D> entity, dynamic value){
  if(value == null) return null;
  if(value is Map){
    return entity.toEntity(value);
  }
  return entity.toEntity(jsonDecode(value));
}

Map<String, dynamic> asMap(dynamic value){
  try{
    if(value is Map){
      return value;
    }
    return jsonDecode(value);
  }catch(e){
    logNUI("NUIEnt", "Failed to decode json of value : $value");
    return null;
  }
}

Map<String, T> asMapOfType<T>(dynamic value){
  final map = asMap(value);
  final mapObject = Map<String, T>();
  map.forEach((key, value) {
    try{
      T data = value as T;
      mapObject[key] = data;
    }catch(e){
      logNUI("NUIEnt", "Failed to map data of ${T.runtimeType} from map");
    }
  });
  return mapObject;
}

Map<String, T> asMapOfObject<T>(NUIEntMapper<T> entity, dynamic value){
  final map = asMap(value);
  final mapObject = Map<String, T>();
  map.forEach((key, value) {
    try{
      T objectData = entity.toEntity(value);
      mapObject[key] = objectData;
    }catch(e){
      logNUI("NUIEnt", "Failed to map object of ${entity.entityName()} from map");
    }
  });
  return mapObject;
}

List<E> asListOfObjects<E>(NUIEntMapper<E> entity, dynamic value){
  if(value == null) return [];
  var endValue = value;
  if(value is String){
    endValue = jsonDecode(value);
  }
  List<E> list = (endValue as List).map((i) =>
      entity.toEntity(i)).toList();
  return list;
}

List<T> asList<T>(dynamic value){
  if(value == null) return null;
  var endValue = value;
  if(value is String){
    endValue = jsonDecode(value);
  }
  return List<T>.from(endValue);
}

DateTime asDateTime(dynamic value, {String format}){
  if(isNullOrEmpty(value)) return null;
  if(value is String){
    return stringToDate(value, dateFormat: format);
  }
  return DateTime.now();
}

class NUIUIEntBinding{
  NUIEntDisplayType displayType;
  String format;
  bool ellipsis;
}

abstract class NUIEnt<T>{
  NUIEntMapper<T> mapper();

  T fromJson(Map<String, dynamic> json){
    return mapper().toEntity(json);
  }

  Map<String, NUIEntField> _fieldMap(){
    final map = Map<String, NUIEntField>();
    mapper().fields().forEach((element) {
      map[element.name] = element;
    });
    return map;
  }

  Map<String, dynamic> toMap(){
    Map<String, dynamic> map = Map();
    for(NUIEntField field in mapper().fields()){
      map[field.name] = field.runGetter(this);
    }
    return map;
  }

  Map<String, dynamic> toMapQuery({Map<NUIEntType, EntCustomerMapper> customMapper}){
    final fieldMap = _fieldMap();
    final json = toMap();
    for(String key in json.keys){
      final val = json[key];
      if(val is NUIEnt){
        json[key] = jsonEncode(val);
      }
      if(val is Map){
        json[key] = jsonEncode(val);
      }
      if(val is List<NUIEnt>){
        json[key] = jsonEncode(val);
      }
      if(val is List){
        json[key] = jsonEncode(val);
      }
      if(val is DateTime){
        final field = fieldMap[key];
        if(field.additionalFormat != null){
          json[key] = dateToString(val, format: field.additionalFormat);
        }
        else {
          json[key] = dateToString(val);
        }
      }

      if(customMapper != null){
        final field = fieldMap[key];
        final mapper = customMapper[field.type];
        if(mapper != null){
          json[key] = mapper(json[key]);
        }
      }
    }
    return json;
  }

  Map<String, dynamic> toJson(){
    final fieldMap = _fieldMap();
    final json = toMap();
    for(String key in json.keys){
      final val = json[key];
      if(val is DateTime){
        final field = fieldMap[key];
        if(field.additionalFormat != null){
          json[key] = dateToString(val, format: field.additionalFormat);
        }
        else {
          json[key] = dateToString(val);
        }
      }
      if(val is NUIEnt){
        json[key] = val.toJson();
      }
      if(val is List<NUIEnt>){
        final mapList = <Map<String, dynamic>>[];
        for(NUIEnt e in val){
          mapList.add(e.toJson());
        }
        json[key] = mapList;
      }
    }
    return json;
  }

  @override
  String toString(){
    return jsonEncode(this);
  }
}

class NUIEntField<T> {
  NUIEntField({@required this.name, @required this.type, @required this.setter, @required this.getter, this.additionalFormat});

  String name;
  NUIEntType type;
  final EntFieldSetter<T> setter;
  final EntFieldGetter<T> getter;
  dynamic additionalFormat;

  void runSetter(T data, dynamic value){
    setter(data, value);
  }

  dynamic runGetter(T data){
    return getter(data);
  }
}

abstract class NUIEntMapper<T>{
  T newInstance();

  String entityName();
  List<NUIEntField> fields();

  T toEntity(Map<String, dynamic> map){
    T data = newInstance();
    for(NUIEntField field in fields()){
//      if(field.type == NUIEntType.OBJECT && field.entity != null){
//        field.runSetter(data, toObject(field.entity, map[field.name]));
//      }
//      else if(field.type == NUIEntType.LIST_OBJECT && field.entity != null){
//        field.runSetter(data, toListOfObjects(field.getTargetEntity(), map[field.name]));
//      }
//      else if(field.type == NUIEntType.LIST_INT){
//        field.runSetter(data, toList<int>(map[field.name]));
//      }
//      else if(field.type == NUIEntType.LIST_STRING){
//        field.runSetter(data, toList<String>(map[field.name]));
//      }
//      else if(field.type == NUIEntType.LIST_DOUBLE){
//        field.runSetter(data, asListOfObjects<double>(field.entity, map[field.name]));
//      }
      try{
        if(field.type == NUIEntType.DATE){
          field.runSetter(data, asDateTime(map[field.name], format: field.additionalFormat));
        }
        else if(field.type == NUIEntType.MAP){
          field.runSetter(data, asMap(map[field.name]));
        }
        else if(field.type == NUIEntType.INTEGER && map[field.name] != null && map[field.name] is int){
          field.runSetter(data, (map[field.name] as int).toInt());
        }
        else if(field.type == NUIEntType.DOUBLE && map[field.name] != null && map[field.name] is double){
          field.runSetter(data, (map[field.name] as double).toDouble());
        }
        else if(field.type == NUIEntType.BOOLEAN){
          final value = map[field.name];
          if(value is String){
            final isTrue = match(value, "true");
            field.runSetter(data, isTrue);
          }
          else if(value is int){
            if(value == 0){
              field.runSetter(data, false);
            }
            else{
              field.runSetter(data, true);
            }
          }
          else{
            field.runSetter(data, value);
          }
        }
        else {
          field.runSetter(data, map[field.name]);
        }
      }catch(e){
        log("NUIJson -> Failed to read field : ${field.name} of type : ${field.type} with exception : ${e.toString()}");
      }

    }
    return data;
  }
}

