import 'package:flutter/cupertino.dart';
import 'package:nui_core/code/nui_code_util.dart';
import 'package:nui_core/config/nui_global_config.dart';
import 'package:nui_core/data/nui_data_util.dart';
import 'package:nui_core/nui_core.dart';

//Data class
class NUIMTLanguage{
  NUIMTLanguage({this.code, this.assetFilename, this.dictionary});
  String code;
  String assetFilename;
  Map<String, String> dictionary;
}

String mtl(String code){
  return NUIMTLingual.get().lang(code);
}


//Singleton class
class NUIMTLingual {
  static NUIMTLingual _instance;
  static const _MODULE = "NUIMTLingual";
  String _currentSelectedLang;

  Map<String, NUIMTLanguage> _languageMapping;

  static NUIMTLingual get(){
    final instance = _instance;
    if(instance == null){
      throwNUI(_MODULE, "Instance have not been built with NUIBusEventBuilder");
    }
    return instance;
  }

  Future changeLang(String lang) async{
    if(!_languageMapping.containsKey(lang)) {
      throwNUI(_MODULE, "Language $lang is not found");
      return;
    }

    _currentSelectedLang = lang;
    await NUISharedPref.get().save("$_MODULE SELECTED_LANG", lang);
    NUIBusEvent.get(module: _MODULE).send(NUIBusEventData<String>(data: lang, sender: _MODULE, code: "Language Change"));
  }

  NUIMTLingual.builder(NUIMTLingualBuilder builder){
    _currentSelectedLang = builder._initialLang;
    _languageMapping = Map();
    for(var l in builder._languages){
      _languageMapping[l.code] = l;
      _currentSelectedLang = _currentSelectedLang ?? l.code;
    }
    _instance = this;
  }

  Future<String> _getSelectedLanguage() async{
    final sharedPref = await NUISharedPrefBuilder().build();
    return await sharedPref.getString("$_MODULE SELECTED_LANG");
  }

  Future _init() async{
    NUIBusEventBuilder().module(_MODULE).build();
    final curLang = await _getSelectedLanguage();
    if(!isNullOrEmpty(curLang)){
      _currentSelectedLang = curLang;
    }

    for(var l in _languageMapping.values){
      if(!isNullOrEmpty(l.assetFilename)){
        await NUIGlobalConfigBuilder().fileName(l.assetFilename).module("$_MODULE ${l.code}").build();
      }
      else if(l.dictionary != null){
        final config = await NUIGlobalConfigBuilder().module("$_MODULE ${l.code}").build();
        config.loadFromMap(l.dictionary);
      }
      continue;
    }
  }

  String get selectedLang{
    return _currentSelectedLang;
  }

  List<String> get availableLangs{
    return _languageMapping.keys.toList();
  }

  String lang(String code){
    if(!_languageMapping.containsKey(_currentSelectedLang)) return null;
    return NUIGlobalConfig.get(module: "$_MODULE $_currentSelectedLang").getString(code);
  }
}


//Build class
class NUIMTLingualBuilder{
  List<NUIMTLanguage> _languages;
  String _initialLang;

  NUIMTLingualBuilder languages(List<NUIMTLanguage> languages){
    this._languages = languages;
    return this;
  }

  NUIMTLingualBuilder initialLanguage(String initialLanguage){
    this._initialLang = initialLanguage;
    return this;
  }

  Future<NUIMTLingual> build() async{
    final mt = NUIMTLingual.builder(this);
    await mt._init();
    return mt;
  }
}


typedef NUIMTLingualWidgetBuilder = Widget Function(BuildContext context, NUIMTLingual mtlang, String lang);
class NUIMTLingualContainer extends StatefulWidget {
  const NUIMTLingualContainer({@required this.builder, Key key}) : super(key: key);
  final NUIMTLingualWidgetBuilder builder;

  @override
  _NUIMTLingualContainerState createState() => _NUIMTLingualContainerState();
}

class _NUIMTLingualContainerState extends State<NUIMTLingualContainer> implements NUIBusEventListener<String> {
  String _selectedLang;

  @override
  void initState(){
    super.initState();
    _selectedLang = NUIMTLingual.get()._currentSelectedLang;
    NUIBusEvent.get(module: NUIMTLingual._MODULE).listen(this);
  }

  @override
  Widget build(BuildContext context) {
    return widget.builder(context, NUIMTLingual.get(), _selectedLang);
  }

  @override
  void onBusEventReceived(NUIBusEventData<String> payload) {
    logNUI("NUIMTLingualContainer", "Received bus event for ${payload.code} with value ${payload.data}");
    if(match(payload.code, "Language Change")){
      setState(() {
        _selectedLang = payload.data;
      });
    }
  }

  @override
  void dispose(){
    super.dispose();
    NUIBusEvent.get(module: NUIMTLingual._MODULE).dispose(this);
  }
}

